-- The ILU
--
-- Signals:
--   reset: global reset. this signal can be seen as chip reset.
--   clk: global clock signal.
--   ilu_enable: ilu_enable signal.
--   ilu_done: ilu_done signal.
--   bef: 4 bits output value of before function.
--   act: 4 bits output value of active function.
--   aft: 4 bits output value of after function.
--   rel: 4 bits output value of partial relation function.
--   a: two bits from operand literal A.
--   b: two bits from operand literal B.
--   c: two bits of the output literal.
--   and_or: and_or signal
--   prime: prime signal
--   to_mem: toMem signal, one bit of config register.
--   redge: re in the thesis, right edge vector signal.
--   water: w in the thesis, water vector signal.
--   empty: empty signal.
--   write_output: this signal is generated when there is a resultant cube.
--   inc_waddr: this signal used to increase mem address unit by 1.
--   cnt_val: counter value, the output of pseudo-random number decoder.
--
-- ILU consists of the iterative network (IT is the cell) and the control
-- unit of the iterative network (called OCU in the thesis).

library ieee;
use ieee.std_logic_1164.all;
use work.all;

-------------------------------------------------------
-- ILU 
-------------------------------------------------------
entity ilu is
  generic 
   ( NumberOfIT : integer := 4 );  -- The number of ITs
  port 
   ( reset, clk, ilu_enable, prime, to_mem, and_or : std_logic;
     rel, bef, act, aft: in std_logic_vector (0 to 3);
     water, redge: in std_logic_vector (0 to NumberOfIT - 1);
     a, b: in std_logic_vector (0 to (NumberOfIT * 2 - 1));
     c: out std_logic_vector (0 to (NumberOfIT * 2 - 1));
     ilu_done, write_output, inc_waddr, empty : out std_logic;
     cnt_val : out std_logic_vector (3 downto 0));
end;

architecture dataflow of ilu is

component ilu_cu
  port (reset, clk, enable, ready, term, to_mem: in std_logic;
        clear, request, init, write_output, inc_waddr, done: out std_logic);
end component;

component itcell
  port (rel, bef, act, aft: in std_logic_vector (0 to 3);
        a, b: in std_logic_vector (0 to 1);
        and_or, redge, redge_pre, water: in std_logic;
        reset, request, clear, prime: in std_logic;
        nxt,     carry,     conf    : in std_logic;  -- propagation signals
        nxt_nxt, carry_nxt, conf_pre: out std_logic; -- propagation signals
        ready: out std_logic;
        c: out std_logic_vector (0 to 1);
        empty_pre : in std_logic;
        empty_nxt, empty : out std_logic;
        cnt_in : in std_logic_vector (3 downto 0);
        cnt_out : out std_logic_vector (3 downto 0);
        count : out std_logic );
end component;

component pcount_decoder
  port (din  : in std_logic_vector (4 downto 0);
        dout : out std_logic_vector (3 downto 0));
end component;

-- carry, nxt (next in the thesis), conf, empty_carry, nct_carry
-- all are propagational signals of the iterative network.
signal clear, request, ready, init, term : std_logic;
signal carry, nxt : std_logic_vector(1 to (NumberOfIT+1));
signal conf : std_logic_vector(0 to NumberOfIT);
signal empty_carry: std_logic_vector(1 to NumberOfIT+1);
signal subready, subempty: std_logic_vector(1 to NumberOfIT);
signal const_one : std_logic := '1';
signal tempready, tempempty: std_logic_vector(1 to NumberOfIT);
signal subcount : std_logic_vector (1 to NumberOfIT);
signal decoder_in: std_logic_vector (4 downto 0);

type sigNx4 is array (1 to NumberOfIT+1) of std_logic_vector(3 downto 0);
signal cnt_carry: sigNx4;

begin

  cu: ilu_cu port map
      (reset=>reset, clk=>clk, enable=>ilu_enable, ready=>ready,
       term=>term, to_mem=>to_mem, clear=>clear, request=>request,
       init=>init, write_output=>write_output,inc_waddr=>inc_waddr,
       done=>ilu_done );

  carry(1) <= '0';           -- carry(1) is "don't care"
  conf(NumberOfIT) <= '0';   -- conf(NumberOfIT) is "don't care"
  empty_carry(1) <= '0';     -- empty_carry(1) is "don't care"

  nxt(1) <= init;
  term <= nxt(NumberOfIT+1);

  cnt_carry(1) <= "1111";

  it: for i in 1 to NumberOfIT generate
    it1: if i=1 generate
      it1: itcell port map
       ( rel=>rel, bef=>bef, act=>act, aft=>aft,
         a=>a(0 to 1), b=>b(0 to 1),
         and_or=>and_or, redge_pre=>const_one, redge=>redge(0), water=>water(0),
         reset=>reset, request=>request, clear=>clear, prime=>prime,
         nxt=>nxt(1),  carry=>carry(1), conf_pre=>conf(0),
         nxt_nxt=>nxt(2), carry_nxt=>carry(2), conf=>conf(1),
         ready=>subready(1), c=>c(0 to 1),
         empty_pre=>empty_carry(1), empty_nxt=>empty_carry(2),
         empty=>subempty(1), count=>subcount(1),
         cnt_in=>cnt_carry(1), cnt_out=>cnt_carry(2)
       );
    end generate it1;

    iti: if i>1 generate
      iti: itcell port map
       ( rel=>rel, bef=>bef, act=>act, aft=>aft,
         a=>a(2*i-2 to 2*i-1), b=>b(2*i-2 to 2*i-1), and_or=>and_or, 
         redge_pre=>redge(i-2), redge=>redge(i-1), water=>water(i-1),
         reset=>reset, request=>request, clear=>clear, prime=>prime,
         nxt=>nxt(i),  carry=>carry(i), conf_pre=>conf(i-1),
         nxt_nxt=>nxt(i+1), carry_nxt=>carry(i+1), conf=>conf(i),
         ready=>subready(i), c=>c(2*i-2 to 2*i-1),
         empty_pre=>empty_carry(i), empty_nxt=>empty_carry(i+1),
         empty=>subempty(i), count=>subcount(i),
         cnt_in=>cnt_carry(i), cnt_out=>cnt_carry(i+1)
       );
    end generate iti;
  end generate it;

  -- the count signal of last IT is combined with 4-bit pseudo-random number.
  -- By using 4-bit counter instead of 5-bit counter, we save 1 iterative 
  -- signal. The decoder of this counter is described by pcountd.vhd
  decoder_in <= subcount(NumberOfIT) & cnt_carry(NumberOfIT);

  -- see pcountd.vhd
  decoder : pcount_decoder port map
   ( din=>decoder_in, dout=>cnt_val );

  -- ready = subready(1) or subready(2) or ... or subready(NumberOfIT)
  tempready(1) <= subready(1);
  readys: for i in 2 to NumberOfIT generate
    tempready(i) <= tempready(i-1) or subready(i);
  end generate readys;
  ready <= tempready(NumberOfIT) after 3 ns;

  -- empty = subempty(1) or subempty(2) or ... or subempty(NumberOfIT)
  tempempty(1) <= subempty(1);
  emptys: for i in 2 to NumberOfIT generate
    tempempty(i) <= tempempty(i-1) or subempty(i);
  end generate emptys;
  empty <= tempempty(NumberOfIT) after 3 ns;

end dataflow;

